#define _GNU_SOURCE
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <limits.h>
#include <sepol/sepol.h>
#include <sepol/policydb/services.h>
#include <selinux/selinux.h>

#define AVCPREFIX "avc:  denied  { "
#define SCONTEXT "scontext="
#define TCONTEXT "tcontext="
#define TCLASS "tclass="

void usage(char *progname, int rc)
{
	fprintf(stderr, "usage:  %s [-p policy] < /var/log/audit/audit.log\n",
		progname);
	exit(rc);
}

struct bool_t {
	const sepol_bool_t * boolean;
	char *name;
	int active;
};

static struct bool_t **boollist = NULL;
static int boolcnt = 0;

struct access_t {
	sepol_handle_t *handle;
	sepol_policydb_t *policydb;
	sepol_security_id_t ssid;
	sepol_security_id_t  tsid;
	sepol_security_class_t tclass;
	sepol_access_vector_t av;
};

static int load_booleans (const sepol_bool_t * boolean,
			  void *arg __attribute__ ((__unused__)) ) {
	boollist[boolcnt] = (struct bool_t *) malloc(sizeof (struct bool_t));
	boollist[boolcnt]->boolean = boolean;
	boollist[boolcnt]->name = strdup(sepol_bool_get_name(boolean));
	boollist[boolcnt]->active = sepol_bool_get_value(boolean);
	boolcnt++;
	return 0;
}

static int check_booleans (struct access_t *access) {
	struct sepol_av_decision avd;
	unsigned int reason;
	int rc;
	int i;
	sepol_bool_key_t *key=NULL;
	int fcnt = 0;
	int *foundlist = calloc(boolcnt, sizeof(int));
	if (!foundlist) {
		fprintf(stderr,
			"Out of memory.\n");
		return fcnt;
	}
	for (i=0; i < boolcnt; i++) {
		char *name = boollist[i]->name;
		int active = boollist[i]->active;
		sepol_bool_t * boolean = (sepol_bool_t *) boollist[i]->boolean;
		rc = sepol_bool_key_create(access->handle,
					   name, 
					   &key);
		if (rc < 0) {
			fprintf(stderr,
				"Could not create boolean key.\n");
			break;
		}
		sepol_bool_set_value(boolean, !active);

		rc = sepol_bool_set(access->handle,
				    access->policydb,
				    key,
				    boolean);
		if (rc < 0) {
			fprintf(stderr,
				"Could not set boolean data %s.\n", name);
			break;
		}

		/* Reproduce the computation. */
		rc = sepol_compute_av_reason(access->ssid, access->tsid, access->tclass, access->av, &avd, &reason);
		if (rc < 0) {
			fprintf(stderr,
				"Error during access vector computation, skipping...\n");
			break;
		} else {
			if (!reason) {
				foundlist[fcnt] = i;
				fcnt++;
			}
			sepol_bool_set_value((sepol_bool_t*)boolean, active);
			rc = sepol_bool_set(access->handle,
					    access->policydb,
					    key,
					    (sepol_bool_t*) boolean);
			if (rc < 0) {
				fprintf(stderr,
					"Could not set boolean data %s.\n", name);
				break;
			}
		}
		sepol_bool_key_free(key);
		key=NULL;		
	}
	if (key)
		sepol_bool_key_free(key);

	if (fcnt > 0)  {
		printf("\tA boolean being set incorrectly.\n");
		for (i = 0; i < fcnt; i++) {
			int ctr = foundlist[i];
			char *name = boollist[ctr]->name;
			int active = boollist[ctr]->active;
			printf("\n\tBoolean %s is %d.\n\tExecute the following to allow access:\n", name, active);
			printf("\t# setsebool -P %s %d\n", name, !active);
		}
	}

	free(foundlist);
	return fcnt;
}


int main(int argc, char **argv)
{
	char path[PATH_MAX];
	char *buffer = NULL, *bufcopy = NULL;
	unsigned int lineno = 0, cnt;
	size_t len = 0, bufcopy_len = 0;
	FILE *fp, *avcp=stdin;
	int opt, rc, set_path = 0;
	char *p, *scon, *tcon, *tclassstr, *permstr;
	sepol_security_id_t ssid, tsid;
	sepol_security_class_t tclass;
	sepol_access_vector_t perm, av;
	struct access_t access;
	struct sepol_av_decision avd;
	unsigned int reason;
	int vers = 0;
	sidtab_t sidtab;
	policydb_t policydb;
	struct policy_file pf;
	
	while ((opt = getopt(argc, argv, "i:p:?h")) > 0) {
		switch (opt) {
		case 'i':
			avcp = fopen(optarg, "r");
			if (!avcp) {
				fprintf(stderr, "%s:  unable to open %s:  %s\n",
					argv[0], path, strerror(errno));
				exit(1);
			}
			break;
			
		case 'p':
			set_path = 1;
			strncpy(path, optarg, PATH_MAX);
			fp = fopen(path, "r");
			if (!fp) {
				fprintf(stderr, "%s:  unable to open %s:  %s\n",
					argv[0], path, strerror(errno));
				exit(1);
			}
			break;
		default:
			usage(argv[0], 0);
		}
	}

	if (argc - optind)
		usage(argv[0], 1);

	if (!set_path) {
		if (!is_selinux_enabled()) {
			fprintf(stderr,
				"%s:  Must specify -p policy on non-SELinux systems\n",
				argv[0]);
			exit(1);
		}
		vers = security_policyvers();
		if (vers < 0) {
			fprintf(stderr,
				"%s:  Could not get policy version:  %s\n",
				argv[0], strerror(errno));
			exit(1);
		}
		snprintf(path, PATH_MAX, "%s.%d",
			 selinux_binary_policy_path(), vers);
		fp = fopen(path, "r");
		while (!fp && errno == ENOENT && --vers) {
			snprintf(path, PATH_MAX, "%s.%d",
				 selinux_binary_policy_path(), vers);
			fp = fopen(path, "r");
		}
		if (!fp) {
			snprintf(path, PATH_MAX, "%s.%d",
				 selinux_binary_policy_path(),
				 security_policyvers());
			fprintf(stderr, "%s:  unable to open %s:  %s\n",
				argv[0], path, strerror(errno));
			exit(1);
		}
	}

	/* Set up a policydb directly so that we can mutate it later
	   for booleans and user settings.  Otherwise we would just use
	   sepol_set_policydb_from_file() here. */
	pf.fp = fp;
	pf.type = PF_USE_STDIO;
	if (policydb_init(&policydb)) {
		fprintf(stderr, "%s:  policydb_init failed: %s\n",
			argv[0], strerror(errno));
		exit(1);
	}
	if (policydb_read(&policydb, &pf, 0)) {
		fprintf(stderr, "%s:  invalid binary policy %s\n",
			argv[0], path);
		exit(1);
	}
	fclose(fp);
	sepol_set_policydb(&policydb);
	if (!set_path) {
		/* If they didn't specify a full path of a binary policy file,
		   then also try loading any boolean settings and user
		   definitions from the active locations.  Otherwise,
		   they can use genpolbools and genpolusers to build a
		   binary policy file that includes any desired settings
		   and then apply audit2why -p to the resulting file. 
		   Errors are non-fatal as such settings are optional. */
		sepol_debug(0);
		(void)sepol_genbools_policydb(&policydb,
					      selinux_booleans_path());
		(void)sepol_genusers_policydb(&policydb, selinux_users_path());
	}

	access.handle = sepol_handle_create();
	access.policydb = (sepol_policydb_t *) &policydb, 

	rc = sepol_bool_count(access.handle,
			      access.policydb, 
			      &cnt); 
	if (rc < 0) {
		fprintf(stderr, "%s:  unable to get bool count\n", argv[0]);
		exit(1);
	}

	boollist = calloc(cnt, sizeof(struct bool_t));
	if (!boollist) {
		fprintf(stderr, "%s:  Out of memory\n", argv[0]);
		exit(1);
	}


	sepol_bool_iterate(access.handle,
			   (const sepol_policydb_t *) &policydb, 
			   load_booleans, 
			   (void *)NULL);


	/* Initialize the sidtab for subsequent use by sepol_context_to_sid
	   and sepol_compute_av_reason. */
	rc = sepol_sidtab_init(&sidtab);
	if (rc < 0) {
		fprintf(stderr, "%s:  unable to init sidtab\n", argv[0]);
		exit(1);
	}
	sepol_set_sidtab(&sidtab);

	/* Process the audit messages. */
	while (getline(&buffer, &len, avcp) > 0) {
		size_t len2 = strlen(buffer);
		char *begin, *end, *search_buf;
		int slen = 0;

		if (buffer[len2 - 1] == '\n')
			buffer[len2 - 1] = 0;
		lineno++;

		p = buffer;
		while (*p && strncmp(p, AVCPREFIX, sizeof(AVCPREFIX) - 1))
			p++;
		if (!(*p))
			continue;	/* not an avc denial */

		p += sizeof(AVCPREFIX) - 1;

		/* Save a copy of the original unmodified buffer. */
		if (!bufcopy) {
			/* Initial allocation */
			bufcopy_len = len;
			bufcopy = malloc(len);
		} else if (bufcopy_len < len) {
			/* Grow */
			bufcopy_len = len;
			bufcopy = realloc(bufcopy, len);
		}
		if (!bufcopy) {
			fprintf(stderr, "%s:  OOM on buffer copy\n", argv[0]);
			exit(2);
		}
		memcpy(bufcopy, buffer, len);

		/* Remember where the permission list begins,
		   and terminate the list. */
		permstr = p;
		while (*p && *p != '}')
			p++;
		if (!(*p)) {
			fprintf(stderr,
				"Missing closing bracket on line %u, skipping...\n",
				lineno);
			continue;
		}
		*p++ = 0;

		search_buf = p;
		/* Get scontext and convert to SID. */
		while (*p && strncmp(p, SCONTEXT, sizeof(SCONTEXT) - 1))
			p++;
		if (!(*p)) {
			fprintf(stderr, "Missing %s on line %u, skipping...\n",
				SCONTEXT, lineno);
			continue;
		}
		p += sizeof(SCONTEXT) - 1;
		begin = p;
		while (*p && !isspace(*p))
			p++;
		end = p;
		slen=end - begin;
		scon = calloc(slen+1, 1);
		strncpy(scon, begin, slen);

		rc = sepol_context_to_sid(scon, strlen(scon) + 1, &ssid);
		if (rc < 0) {
			fprintf(stderr,
				"Invalid %s%s on line %u, skipping...\n",
				SCONTEXT, scon, lineno);
			continue;
		}

		free(scon);
		/* start searching at the beginning again */
		p = search_buf;

		/* Get tcontext and convert to SID. */
		while (*p && strncmp(p, TCONTEXT, sizeof(TCONTEXT) - 1))
			p++;
		if (!(*p)) {
			fprintf(stderr, "Missing %s on line %u, skipping...\n",
				TCONTEXT, lineno);
			continue;
		}
		p += sizeof(TCONTEXT) - 1;

		begin = p;
		while (*p && !isspace(*p))
			p++;
		end = p;
		slen=end - begin;
		tcon = calloc(slen+1, 1);
		strncpy(tcon, begin, slen);

		rc = sepol_context_to_sid(tcon, strlen(tcon) + 1, &tsid);
		if (rc < 0) {
			fprintf(stderr,
				"Invalid %s%s on line %u, skipping...\n",
				TCONTEXT, tcon, lineno);
			continue;
		}
		free(tcon);
		/* start searching at the beginning again */
		p = search_buf;

		/* Get tclass= and convert to value. */
		while (*p && strncmp(p, TCLASS, sizeof(TCLASS) - 1))
			p++;
		if (!(*p)) {
			fprintf(stderr, "Missing %s on line %u, skipping...\n",
				TCLASS, lineno);
			continue;
		}
		p += sizeof(TCLASS) - 1;
		begin = p;
		while (*p && !isspace(*p))
			p++;

		end = p;
		slen=end - begin;
		tclassstr = calloc(slen+1, 1);
		strncpy(tclassstr, begin, slen);

		tclass = string_to_security_class(tclassstr);
		free(tclassstr);
		if (!tclass) {
			fprintf(stderr,
				"Invalid %s%s on line %u, skipping...\n",
				TCLASS, tclassstr, lineno);
			continue;
		}

		/* Convert the permission list to an AV. */
		p = permstr;
		av = 0;
		while (*p) {
			while (*p && !isspace(*p))
				p++;
			if (*p)
				*p++ = 0;
			perm = string_to_av_perm(tclass, permstr);
			if (!perm) {
				fprintf(stderr,
					"Invalid permission %s on line %u, skipping...\n",
					permstr, lineno);
				continue;
			}
			av |= perm;
			permstr = p;
		}

		/* Reproduce the computation. */
		rc = sepol_compute_av_reason(ssid, tsid, tclass, av, &avd,
					     &reason);
		if (rc < 0) {
			fprintf(stderr,
				"Error during access vector computation on line %u, skipping...\n",
				lineno);
			continue;
		}

		printf("%s\n\tWas caused by:\n", bufcopy);

		if (!reason) {
			printf("\t\tUnknown - would be allowed by %s policy\n",
			       set_path ? "specified" : "active");
			printf
			    ("\t\tPossible mismatch between this policy and the one under which the audit message was generated.\n");
			printf
			    ("\t\tPossible mismatch between current in-memory boolean settings vs. permanent ones.\n");
		}

		if (reason & SEPOL_COMPUTEAV_TE) {
			access.ssid = ssid;
			access.tsid = tsid;
			access.tclass = tclass;
			access.av = av;
			
			if (check_booleans(&access) == 0) {
				printf("\t\tMissing or disabled TE allow rule.\n");
				printf
					("\t\tYou can see the necessary allow rules by running audit2allow with this audit message as input.\n");
			}
		}

		if (reason & SEPOL_COMPUTEAV_CONS) {
			printf("\t\tConstraint violation.\n");
			printf("\t\tCheck policy/constraints.\n");
			printf
			    ("\t\tTypically, you just need to add a type attribute to the domain to satisfy the constraint.\n");
		}

		if (reason & SEPOL_COMPUTEAV_RBAC) {
			printf("\t\tMissing role allow rule.\n");
			printf("\t\tAdd allow rule for the role pair.\n");
		}

		printf("\n");
	}
	free(buffer);
	free(bufcopy);
	if (avcp != stdin)
		fclose(avcp);

	exit(0);
}
